import cv2
import numpy as np
import copy
import csv

search_order = [
    [7, 0, 1, 2, 3, 4, 5, 6],
    [7, 0, 1, 2, 3, 4, 5, 6],
    [1, 2, 3, 4, 5, 6, 7, 0],
    [1, 2, 3, 4, 5, 6, 7, 0],
    [3, 4, 5, 6, 7, 0, 1, 2],
    [3, 4, 5, 6, 7, 0, 1, 2],
    [5, 6, 7, 0, 1, 2, 3, 4],
    [5, 6, 7, 0, 1, 2, 3, 4]
]
move_x = [1, 1, 0, -1, -1, -1, 0, 1]
move_y = [0, -1, -1, -1, 0, 1, 1, 1]

initial_direction_table = [0, 3, 2, 3, 1, 3, 2, 3, 4, 4, 4, 4, 4, 4, 4, 4]

image_size_max_x = 3000
image_size_max_y = 3000
chain_code_data_max = 50000


class Chaincode:
    chain_position_x = [0]*image_size_max_x
    chain_position_y = [0]*image_size_max_y
    chain_position_yx = []
    chaincode = [0]*chain_code_data_max
    chaincode_length = 0
    x_min = 0
    y_min = 0
    x_max = 0
    y_max = 0


class Parameters:
    pixel_count = 0


d_table = [
    [0, 1, 1, 1, 1, 0, 0, 0],
    [0, 1, 1, 1, 1, 0, 0, 0],
    [0, 1, 1, 1, 1, 0, 0, 0],
    [0, 1, 1, 1, 1, 0, 0, 0],
    [-1, 0, 0, 0, 0, -1, -1, -1],
    [-1, 0, 0, 0, 0, -1, -1, -1],
    [-1, 0, 0, 0, 0, -1, -1, -1],
    [-1, 0, 0, 0, 0, -1, -1, -1]
]

c_table = [
    [0, 0, 0, 0, 0, 0, 0, 0],
    [0, 0, 0, 0, 0, 1, 0, 0],
    [0, 0, 0, 0, 0, 1, 1, 0],
    [0, 0, 0, 0, 0, 1, 1, 1],
    [1, 0, 0, 0, 0, 1, 1, 1],
    [1, 1, 0, 0, 0, 1, 1, 1],
    [1, 1, 1, 0, 0, 1, 1, 1],
    [1, 1, 1, 1, 0, 1, 1, 1]
]


im_rgb = cv2.imread("data/Particle.jpg")
print(im_rgb.shape)
# (225, 400, 3)
print(im_rgb.dtype)

im_gray = cv2.cvtColor(im_rgb, cv2.COLOR_BGR2GRAY)
print(im_gray.shape)
# (225, 400)

print(im_gray.dtype)
# uint8

# rtn = cv2.imwrite("lena_gray.jpg", im_gray)
# if rtn:
#     print('True')
# else:
#     print('False')


def image_to_csv(img):
    row_size = img.shape[0]
    column_size = img.shape[1]

    # 輝度値をCSVファイルで保存
    with open('image_data.csv', 'w', newline='') as csvfile:
        writer = csv.writer(csvfile)
        writer.writerows(img)


def get_threshold_level(gray_img, ratio):
    row_size = gray_img.shape[0]
    column_size = gray_img.shape[1]
    # 輝度の濃淡レベルは8bit（0-255：256段階）とする
    histogram = [0]*256
    _threshold_level = 0

    # 輝度ヒストグラムの作成
    for row in range(row_size):
        for column in range(column_size):
            histogram[gray_img[row][column]] += 1

    # 最頻度輝度値の算出
    mode = histogram.index(max(histogram))
    _threshold_level = int(mode * 0.01 * ratio + 0.5)

    return _threshold_level


def binarizer(gray_img, threshold):
    row_size = gray_img.shape[0]
    column_size = gray_img.shape[1]
    # 空の二次元配列を作成
    bin_img = np.empty((row_size, column_size))
    # 輝度値が閾値より暗ければ（小さければ）1に変換、そうでなければ0に変換し2値化する
    for row in range(row_size):
        for column in range(column_size):
            if gray_img[row][column] > threshold:
                bin_img[row][column] = 0
            else:
                bin_img[row][column] = 1
    return bin_img


def raster_scan(bin_img):
    _result = False
    _start_point_row = 0
    _start_point_column = 0
    row_size = bin_img.shape[0]
    column_size = bin_img.shape[1]
    # 最初に1が登場する画素の位置を探して関数から抜ける。位置情報を返す。
    for row in range(row_size):
        for column in range(column_size):
            if bin_img[row][column] == 1:
                _start_point_row = row
                _start_point_column = column
                _result = True
                return _result, _start_point_row, _start_point_column


def get_initial_direction(start_y, start_x, bin_img):
    initial_direction = 0
    _current_x = start_x
    _current_y = start_y
    # 初期方向コードを得るための輝度データを取得
    b4 = int(bin_img[_current_y][_current_x])
    b3 = int(bin_img[_current_y][_current_x + 1])
    b2 = int(bin_img[_current_y + 1][_current_x - 1])
    b1 = int(bin_img[_current_y + 1][_current_x])
    b0 = int(bin_img[_current_y + 1][_current_x + 1])
    if b4 == 0:
        initial_direction = 0
        return initial_direction
    code = (b3 << 3) + (b2 << 2) + (b1 << 1) + b0
    initial_direction = initial_direction_table[code]
    return initial_direction


def making_chain_code(start_y, start_x, initial_direction, chaincode_reg, bin_img):
    _current_x = start_x
    _current_y = start_y
    _first_direction = initial_direction
    _around_binary_data = [0]*8
    chaincode_reg.chain_position_x[0] = _current_x
    chaincode_reg.chain_position_y[0] = _current_y
    chaincode_reg.chain_position_yx.append([chaincode_reg.chain_position_y[0], chaincode_reg.chain_position_x[0]])
    chaincode_reg.chaincode[0] = initial_direction

    while True:
        # 注目画素の8近傍画素輝度データを取得
        _around_binary_data[0] = bin_img[_current_y][_current_x + 1]
        _around_binary_data[1] = bin_img[_current_y - 1][_current_x + 1]
        _around_binary_data[2] = bin_img[_current_y - 1][_current_x]
        _around_binary_data[3] = bin_img[_current_y - 1][_current_x - 1]
        _around_binary_data[4] = bin_img[_current_y][_current_x - 1]
        _around_binary_data[5] = bin_img[_current_y + 1][_current_x - 1]
        _around_binary_data[6] = bin_img[_current_y + 1][_current_x]
        _around_binary_data[7] = bin_img[_current_y + 1][_current_x + 1]
        # 次に進むべき画素の位置を探す
        for i in range(8):
            direction = search_order[_first_direction][i]
            if _around_binary_data[direction] == 1:
                _current_x = _current_x + move_x[direction]
                _current_y = _current_y + move_y[direction]
                chaincode_reg.chaincode[chaincode_reg.chaincode_length] = direction
                _first_direction = direction
                chaincode_reg.chaincode_length += 1
                if chaincode_reg.chaincode_length > chain_code_data_max:
                    return False
                chaincode_reg.chain_position_x[chaincode_reg.chaincode_length] = _current_x
                chaincode_reg.chain_position_y[chaincode_reg.chaincode_length] = _current_y
                chaincode_reg.chain_position_yx.append(
                    [chaincode_reg.chain_position_y[chaincode_reg.chaincode_length],
                     chaincode_reg.chain_position_x[chaincode_reg.chaincode_length]])
                break
        chaincode_reg.x_min = min(chaincode_reg.chain_position_x[:chaincode_registers.chaincode_length])
        chaincode_reg.y_min = min(chaincode_reg.chain_position_y[:chaincode_registers.chaincode_length])
        chaincode_reg.x_max = max(chaincode_reg.chain_position_x[:chaincode_registers.chaincode_length])
        chaincode_reg.y_max = max(chaincode_reg.chain_position_y[:chaincode_registers.chaincode_length])
        if (_current_x == start_x) and (_current_y == start_y):
            break
    return True


def edge_paint(chain_yx, img_rgb):
    new_img = copy.copy(img_rgb)
    for yx in chain_yx:
        new_img[yx[0]][yx[1]] = (0, 0, 255)
    return new_img


def calculate_image_parameter(chaincode_reg, parameters):
    _pixel = 0
    x = d = c = 0
    # chaincode_reg.chain_position_x[chaincode_reg.chaincode_length] = chaincode_reg.chain_position_x[0]
    chaincode_reg.chaincode[chaincode_reg.chaincode_length] = chaincode_reg.chaincode[0]
    for i in range(1, chaincode_reg.chaincode_length+1):
        x = chaincode_reg.chain_position_x[i]
        d = d_table[chaincode_reg.chaincode[i-1]][chaincode_reg.chaincode[i]]
        c = c_table[chaincode_reg.chaincode[i-1]][chaincode_reg.chaincode[i]]
        _pixel += (x * d + c)
    parameters.pixel_count = _pixel


# 2値化ための閾値を得る
threshold_level = get_threshold_level(im_gray, 90)
print('threshold_level is ', threshold_level)
# グレースケールのRAW画像を2値化する。
bin_im_gray = binarizer(im_gray, threshold_level)


# 2値化した画像をcsvファイルで出力
image_to_csv(im_gray)

# ラスタースキャンを行い輪郭抽出処理のスタート座標を得る。
raster_scan_result, start_point_row, start_point_column = raster_scan(bin_im_gray)
if raster_scan_result:
    print(start_point_row, start_point_column)

# 輪郭追跡のための初期方向コードを得る。
initial_direction_code = get_initial_direction(start_point_row, start_point_column, bin_im_gray)
if initial_direction_code:
    print('initial_direction_code is ', initial_direction_code)

chaincode_registers = Chaincode()
chaincode_result = making_chain_code(start_point_row, start_point_column, initial_direction_code, chaincode_registers, bin_im_gray)
if chaincode_result:
    print('chaincode_length is ', chaincode_registers.chaincode_length)
    print('chain_position_x is ', chaincode_registers.chain_position_x[:chaincode_registers.chaincode_length])
    print('chain_position_y is ', chaincode_registers.chain_position_y[:chaincode_registers.chaincode_length])
    print('x_min is ', chaincode_registers.x_min)
    print('y_min is ', chaincode_registers.y_min)
    print('x_max is ', chaincode_registers.x_max)
    print('y_max is ', chaincode_registers.y_max)
    print('chain_position_yx is ', chaincode_registers.chain_position_yx[:chaincode_registers.chaincode_length])
    print('chaincode is ', chaincode_registers.chaincode[:chaincode_registers.chaincode_length])


edge_img = edge_paint(chaincode_registers.chain_position_yx, im_rgb)

rtn = cv2.imwrite("edge_img.bmp", edge_img)
if rtn:
    print('True')
else:
    print('False')

image_parameters = Parameters()
calculate_image_parameter(chaincode_registers, image_parameters)
print('pixel_count is ', image_parameters.pixel_count)
